@@ -7,19 +7,19 @@ This module implements (partially) the W3C recommendation
 
 .. seealso:: https://www.w3.org/TR/tabular-metadata/
 """
-
-from __future__ import unicode_literals
-
+import io
 import re
 import json
 import shutil
+import pathlib
+import zipfile
 import operator
 import warnings
 import itertools
+import contextlib
 import collections
-
-from ._compat import (pathlib, text_type, iteritems, itervalues, zip,
-    py3_unicode_to_str, json_open, urljoin)
+from urllib.parse import urljoin, urlparse, urlunparse
+from urllib.request import urlopen
 
 import attr
 import uritemplate
@@ -27,6 +27,7 @@ import uritemplate
 from . import utils
 from .datatypes import DATATYPES
 from .dsv import Dialect, UnicodeReaderWithLineNumber, UnicodeWriter
+from .frictionless import DataPackage
 
 DEFAULT = object()
 
@@ -35,22 +36,27 @@ __all__ = [
     'Table', 'Column', 'ForeignKey',
     'Link', 'NaturalLanguage',
     'Datatype',
+    'is_url',
 ]
 
 
-# Level 1 variable names according to https://tools.ietf.org/html/rfc6570#section-2.3:
-_varchar = '([a-zA-Z0-9_]|\%[a-fA-F0-9]{2})'
-_varname = re.compile('(' + _varchar + '([.]?' + _varchar + ')*)$')
+def is_url(s):
+    return re.match(r'https?://', str(s))
 
 
-def log_or_raise(msg, log=None, level='warn', exception_cls=ValueError):
+def log_or_raise(msg, log=None, level='warning', exception_cls=ValueError):
     if log:
         getattr(log, level)(msg)
     else:
         raise exception_cls(msg)
 
 
-def nolog(level='warn'):
+def json_open(filename, mode='r', encoding='utf-8'):
+    assert encoding == 'utf-8'
+    return io.open(filename, mode, encoding=encoding)
+
+
+def nolog(level='warning'):
     from types import MethodType
 
     class Log(object):
@@ -86,7 +92,6 @@ def uri_template_property():
         converter=lambda v: v if v is None else URITemplate(v))
 
 
-@py3_unicode_to_str
 class Link(object):
     """
 
@@ -96,7 +101,7 @@ class Link(object):
     def __init__(self, string):
         self.string = string
 
-    def __unicode__(self):
+    def __str__(self):
         return self.string
 
     def asdict(self, omit_defaults=True):
@@ -126,7 +131,6 @@ def link_property():
         converter=lambda v: v if v is None else Link(v))
 
 
-@py3_unicode_to_str
 class NaturalLanguage(collections.OrderedDict):
     """
 
@@ -136,12 +140,12 @@ class NaturalLanguage(collections.OrderedDict):
     def __init__(self, value):
         super(NaturalLanguage, self).__init__()
         self.value = value
-        if isinstance(self.value, text_type):
+        if isinstance(self.value, str):
             self[None] = [self.value]
         elif isinstance(self.value, (list, tuple)):
             self[None] = list(self.value)
         elif isinstance(self.value, dict):
-            for k, v in iteritems(self.value):
+            for k, v in self.value.items():
                 if not isinstance(v, (list, tuple)):
                     v = [v]
                 self[k] = v
@@ -155,15 +159,15 @@ class NaturalLanguage(collections.OrderedDict):
             return self[None]
         return collections.OrderedDict(
             ('und' if k is None else k, v[0] if len(v) == 1 else v)
-            for k, v in iteritems(self))
+            for k, v in self.items())
 
     def add(self, string, lang=None):
         if lang not in self:
             self[lang] = []
         self[lang].append(string)
 
-    def __unicode__(self):
-        return self.getfirst() or next(itervalues(self))[0]
+    def __str__(self):
+        return self.getfirst() or next(iter(self.values()))[0]
 
     def getfirst(self, lang=None):
         return self.get(lang, [None])[0]
@@ -173,7 +177,7 @@ class NaturalLanguage(collections.OrderedDict):
 class DescriptionBase(object):
     """Container for
     - common properties (see http://w3c.github.io/csvw/metadata/#common-properties)
-    - @-properies.
+    - @-properties.
     """
 
     common_props = attr.ib(default=attr.Factory(dict))
@@ -182,7 +186,7 @@ class DescriptionBase(object):
     @staticmethod
     def partition_properties(d):
         c, a, dd = {}, {}, {}
-        for k, v in iteritems(d or {}):
+        for k, v in (d or {}).items():
             if k.startswith('@'):
                 a[k[1:]] = v
             elif ':' in k:
@@ -204,20 +208,19 @@ class DescriptionBase(object):
                 return [_asdict_single(vv) for vv in v]
             return _asdict_single(v)
 
-        for k, v in sorted(iteritems(self.at_props)):
+        for k, v in sorted(self.at_props.items()):
             yield '@' + k, _asdict_multiple(v)
 
-        for k, v in sorted(iteritems(self.common_props)):
+        for k, v in sorted(self.common_props.items()):
             yield k, _asdict_multiple(v)
 
-        for k, v in iteritems(utils.attr_asdict(self, omit_defaults=omit_defaults)):
+        for k, v in utils.attr_asdict(self, omit_defaults=omit_defaults).items():
             if k not in ('common_props', 'at_props'):
                 yield k, _asdict_multiple(v)
 
     def asdict(self, omit_defaults=True):
         return collections.OrderedDict(
-            (k, v) for k, v in
-            self._iter_dict_items(omit_defaults) if v not in (None, [], {}))
+            (k, v) for k, v in self._iter_dict_items(omit_defaults) if v not in ([], {}))
 
 
 def optional_int():
@@ -256,7 +259,7 @@ class Datatype(DescriptionBase):
         instance.
         :return: An instance of `cls`
         """
-        if isinstance(v, text_type):
+        if isinstance(v, str):
             return cls(base=v)
 
         if isinstance(v, dict):
@@ -364,12 +367,9 @@ class Description(DescriptionBase):
 
 
 @attr.s
-@py3_unicode_to_str
 class Column(Description):
 
-    name = attr.ib(
-        default=None,
-        validator=utils.attr_valid_re(_varname, nullable=True))
+    name = attr.ib(default=None)
     suppressOutput = attr.ib(default=False)
     titles = attr.ib(
         default=None,
@@ -378,7 +378,7 @@ class Column(Description):
     virtual = attr.ib(default=False)
     _number = attr.ib(default=None, repr=False)
 
-    def __unicode__(self):
+    def __str__(self):
         return self.name or \
             (self.titles and self.titles.getfirst()) or \
             '_col.{}'.format(self._number)
@@ -483,15 +483,26 @@ class Schema(Description):
     rowTitles = attr.ib(default=attr.Factory(list))
 
     def __attrs_post_init__(self):
-        virtual = False
+        virtual, seen = False, set()
         for i, col in enumerate(self.columns):
             if col.virtual:  # first virtual column sets the flag
                 virtual = True
             elif virtual:  # non-virtual column after virtual column!
                 raise ValueError('no non-virtual column allowed after virtual columns')
+            if not virtual:
+                if col.header in seen:
+                    warnings.warn('Duplicate column name!')
+                seen.add(col.header)
             col._parent = self
             col._number = i + 1
 
+    @classmethod
+    def fromvalue(cls, v):
+        if isinstance(v, str):
+            # The schema is referenced with a URL
+            v = json.loads(urlopen(v).read().decode('utf8'))
+        return cls(**cls.partition_properties(v))
+
     @property
     def columndict(self):
         return {c.header: c for c in self.columns}
@@ -520,6 +531,8 @@ class TableLike(Description):
         default=None,
         converter=lambda v: Schema.fromvalue(v))
     transformations = attr.ib(default=attr.Factory(list))
+    url = link_property()
+    _fname = attr.ib(default=None)  # The path of the metadata file.
 
     def __attrs_post_init__(self):
         if self.tableSchema:
@@ -528,25 +541,90 @@ class TableLike(Description):
     def get_column(self, spec):
         return self.tableSchema.get_column(spec) if self.tableSchema else None
 
+    @classmethod
+    def from_file(cls, fname):
+        if not isinstance(fname, pathlib.Path):
+            if is_url(fname):
+                return cls.from_url(str(fname))
+            fname = pathlib.Path(fname)
+        with json_open(str(fname)) as f:
+            data = json.load(f)
+        res = cls.fromvalue(data)
+        res._fname = fname
+        return res
+
+    @classmethod
+    def from_url(cls, url):
+        with io.TextIOWrapper(urlopen(url), encoding='utf8') as f:
+            data = json.load(f)
+        res = cls.fromvalue(data)
+        if 'base' not in res.at_props:
+            url = urlparse(url)
+            res.at_props['base'] = urlunparse((url.scheme, url.netloc, url.path, '', '', ''))
+        return res
+
+    def to_file(self, fname, omit_defaults=True):
+        if not isinstance(fname, pathlib.Path):
+            fname = pathlib.Path(fname)
+        data = self.asdict(omit_defaults=omit_defaults)
+        with json_open(str(fname), 'w') as f:
+            json.dump(data, f, indent=4, separators=(',', ': '))
+        return fname
+
+    @property
+    def base(self):
+        """
+        We only support data in the filesystem, thus we make sure `base` is a `pathlib.Path`.
+        """
+        at_props = self._parent.at_props if self._parent else self.at_props
+        if 'base' in at_props:
+            return at_props['base']
+        return self._parent._fname.parent if self._parent else self._fname.parent
+
 
 @attr.s
 class Table(TableLike):
 
-    url = link_property()
     suppressOutput = attr.ib(default=False)
 
+    def add_foreign_key(self, colref, ref_resource, ref_colref):
+        """
+        Add a foreign key constraint to `tableSchema.foreignKeys`.
+
+        :param colref: Column reference for the foreign key.
+        :param ref_resource: Referenced table.
+        :param ref_colref: Column reference of the key in the referenced table.
+        """
+        colref = [colref] if not isinstance(colref, (tuple, list)) else colref
+        if not all(col in [c.name for c in self.tableSchema.columns] for col in colref):
+            raise ValueError('unknown column in foreignKey {0}'.format(colref))
+
+        self.tableSchema.foreignKeys.append(ForeignKey.fromdict({
+            'columnReference': colref,
+            'reference': {'resource': ref_resource, 'columnReference': ref_colref}
+        }))
+
     @property
     def local_name(self):
-        return self.url.string
+        return self.url.string if self.url else None
 
     def _get_dialect(self):
         return self.dialect or (self._parent and self._parent.dialect) or Dialect()
 
-    def write(self, items, fname=DEFAULT, base=None):
+    def write(self, items, fname=DEFAULT, base=None, _zipped=False):
+        """
+        Write row items to a CSV file according to the table schema.
+
+        :param items: Iterator of `dict`s storing the data per row.
+        :param fname: Name of the file to which to write the data.
+        :param base: Base directory relative to which to interpret table urls.
+        :param _zipped: Flag signaling whether the resulting data file should be zipped.
+        :return: The CSV content if `fname==None` else the number of rows written.
+        """
         dialect = self._get_dialect()
         non_virtual_cols = [c for c in self.tableSchema.columns if not c.virtual]
         if fname is DEFAULT:
-            fname = self.url.resolve(pathlib.Path(base) if base else self._parent.base)
+            fname = self.url.resolve(pathlib.Path(base) if base else self.base)
 
         rowcount = 0
         with UnicodeWriter(fname, dialect=dialect) as writer:
@@ -564,6 +642,11 @@ class Table(TableLike):
                 writer.writerow(row)
             if fname is None:
                 return writer.read()
+        if fname and _zipped:
+            fpath = pathlib.Path(fname)
+            with zipfile.ZipFile(str(fpath.parent.joinpath(fpath.name + '.zip')), 'w') as zipf:
+                zipf.write(str(fpath))
+            fpath.unlink()
         return rowcount
 
     def check_primary_key(self, log=None, items=None):
@@ -597,6 +680,9 @@ class Table(TableLike):
         table cells, or for virtual columns (which have no values) the valueUrl for that
         column. This includes columns not specified in the table specification.
 
+        Note: If the resolved data filename does not exist - but there is a zip file of the form
+        `fname+'.zip'`, we try to read the data from this file after unzipping.
+
         :param log: Logger object (default None) The object that reports parsing errors.\
         If none is given, parsing errors raise ValueError instead.
         :param bool with_metadata: (default False) Also yield fname and lineno
@@ -605,7 +691,7 @@ class Table(TableLike):
         :return: A generator of dicts or triples (fname, lineno, dict) if with_metadata
         """
         dialect = self._get_dialect()
-        fname = fname or self.url.resolve(self._parent.base)
+        fname = fname or self.url.resolve(self.base)
         colnames, virtualcols, requiredcols = [], [], set()
         for col in self.tableSchema.columns:
             if col.virtual:
@@ -616,7 +702,21 @@ class Table(TableLike):
             if col.required:
                 requiredcols.add(col.header)
 
-        with UnicodeReaderWithLineNumber(fname, dialect=dialect) as reader:
+        with contextlib.ExitStack() as stack:
+            if is_url(fname):
+                handle = io.TextIOWrapper(urlopen(str(fname)), encoding=dialect.encoding)
+            else:
+                handle = fname
+                fpath = pathlib.Path(fname)
+                if not fpath.exists():
+                    zipfname = fpath.parent.joinpath(fpath.name + '.zip')
+                    if zipfname.exists():
+                        zipf = stack.enter_context(zipfile.ZipFile(str(zipfname)))
+                        handle = io.TextIOWrapper(
+                            zipf.open([n for n in zipf.namelist() if n.endswith(fpath.name)][0]),
+                            encoding=dialect.encoding)
+
+            reader = stack.enter_context(UnicodeReaderWithLineNumber(handle, dialect=dialect))
             reader = iter(reader)
 
             # If the data file has a header row, this row overrides the header as
@@ -632,6 +732,9 @@ class Table(TableLike):
             # If columns in the data are ordered as in the spec, we can match values to
             # columns by index, rather than looking up columns by name.
             if header == colnames:
+                # Note that virtual columns are only allowed to come **after** regular ones,
+                # so we can simply zip the whole columns list, and silently ignore surplus
+                # virtual columns.
                 header_cols = list(zip(header, self.tableSchema.columns))
             else:
                 header_cols = [(h, self.tableSchema.get_column(h)) for h in header]
@@ -657,6 +760,8 @@ class Table(TableLike):
                                 log=log)
                             error = True
                     else:
+                        warnings.warn(
+                            'Unspecified column "{0}" in table {1}'.format(k, self.local_name))
                         res[k] = v
 
                 for k, j in required.items():
@@ -667,6 +772,10 @@ class Table(TableLike):
                             log=log)
                         error = True
 
+                # Augment result with regular columns not provided in the data:
+                for key in colnames:
+                    res.setdefault(key, None)
+
                 # Augment result with virtual columns:
                 for key, valueUrl in virtualcols:
                     res[key] = valueUrl.expand(**res)
@@ -681,35 +790,19 @@ class Table(TableLike):
 @attr.s
 class TableGroup(TableLike):
 
-    _fname = attr.ib(default=None)  # The path of the metadata file.
-    url = attr.ib(default=None)
     tables = attr.ib(
         repr=False,
         default=attr.Factory(list),
         converter=lambda v: [Table.fromvalue(vv) for vv in v])
 
-    @classmethod
-    def from_file(cls, fname):
-        if not isinstance(fname, pathlib.Path):
-            fname = pathlib.Path(fname)
-        with json_open(str(fname)) as f:
-            data = json.load(f)
-        res = cls.fromvalue(data)
-        res._fname = fname
-        return res
-
     def __attrs_post_init__(self):
         TableLike.__attrs_post_init__(self)
         for table in self.tables:
             table._parent = self
 
-    def to_file(self, fname, omit_defaults=True):
-        if not isinstance(fname, pathlib.Path):
-            fname = pathlib.Path(fname)
-        data = self.asdict(omit_defaults=omit_defaults)
-        with json_open(str(fname), 'w') as f:
-            json.dump(data, f, indent=4, separators=(',', ': '))
-        return fname
+    @classmethod
+    def from_frictionless_datapackage(cls, dp):
+        return DataPackage(dp).to_tablegroup(cls)
 
     def read(self):
         """
@@ -717,7 +810,7 @@ class TableGroup(TableLike):
         """
         return {tname: list(t.iterdicts()) for tname, t in self.tabledict.items()}
 
-    def write(self, fname, **items):
+    def write(self, fname, _zipped=False, **items):
         """
         Write a TableGroup's data and metadata to files.
 
@@ -727,7 +820,7 @@ class TableGroup(TableLike):
         """
         fname = pathlib.Path(fname)
         for tname, rows in items.items():
-            self.tabledict[tname].write(rows, base=fname.parent)
+            self.tabledict[tname].write(rows, base=fname.parent, _zipped=_zipped)
         self.to_file(fname)
 
     def copy(self, dest):
@@ -748,13 +841,6 @@ class TableGroup(TableLike):
     def tabledict(self):
         return {t.local_name: t for t in self.tables}
 
-    @property
-    def base(self):
-        """
-        We only support data in the filesystem, thus we make sure `base` is a `pathlib.Path`.
-        """
-        return self._fname.parent
-
     def check_referential_integrity(self, data=None, log=None):
         success = True
         if data is not None:
@@ -770,7 +856,12 @@ class TableGroup(TableLike):
             if not fk.reference.schemaReference]
         # FIXME: We only support Foreign Key references between tables!
         fkeys = sorted(fkeys, key=lambda x: (x[0].local_name, x[1], x[2].local_name))
-        for table, grp in itertools.groupby(fkeys, lambda x: x[0]):
+        # Grouping by local_name of tables - even though we'd like to have the table objects
+        # around, too. This it to prevent going down the rabbit hole of comparing table objects
+        # for equality, when comparison of the string names is enough.
+        for _, grp in itertools.groupby(fkeys, lambda x: x[0].local_name):
+            grp = list(grp)
+            table = grp[0][0]
             t_fkeys = [(key, [(child, ref) for _, _, child, ref in kgrp])
                        for key, kgrp in itertools.groupby(grp, lambda x: x[1])]
             get_seen = [(operator.itemgetter(*key), set()) for key, _ in t_fkeys]
@@ -798,7 +889,7 @@ class TableGroup(TableLike):
                                 continue
                             elif colref not in seen:
                                 log_or_raise(
-                                    '{0}:{1} Key {2} not found in table {3}'.format(
+                                    '{0}:{1} Key `{2}` not found in table {3}'.format(
                                         fname,
                                         lineno,
                                         colref,
@@ -806,7 +897,3 @@ class TableGroup(TableLike):
                                     log=log)
                                 success = False
         return success
-
-    #
-    # FIXME: to_sqlite()!
-    #
